""" dirwatch example

    Watches one or more directories for new files.
    Simply prints the timestamp and path, when the event is seen.

    eg.
    > python dwexample.py -d c:\\temp -d c:\\foo -d c:\\bar
    Adding watch: c:\\temp
    Adding watch: c:\\foo
    Adding watch: c:\\bar
    Listening (press Ctrl-C to quit)...
    Wed Mar 02 17:27:21 2011 [EV_FILE_CREATED] c:\\temp\\dbtest - Copy.py
    Wed Mar 02 17:28:18 2011 [EV_FILE_CREATED] c:\\temp\\bad_avl.png
    Wed Mar 02 17:28:50 2011 [EV_FILE_CREATED] c:\\temp\\temp.7z
    quitting...
    done.
"""

import argparse
import os
import os.path
import time

import dirwatch
from dirwatch.eventtype import EV_FILE_CREATED

def parse_cmdline():
    parser = argparse.ArgumentParser(description="Watch for new files in given directories.")
    parser.add_argument('-d', '--directory', metavar="DIR_PATH", dest='directories', action='append', help='Monitor new files added to this directory', default=[])
    args = parser.parse_args()
    return args

def on_new_file(event):
    print "%s [%s] %s" % (time.ctime(event.timestamp), event.event_type, event.path)

def monitor_directories(paths, action=None):
    """ Watch the given paths for newly created files.

    Accepts a list of valid directory paths, and a callable
    to handle the event when it happens.

    The callable should accept at least one parameter, which will have the attributes:
       timestamp, path, event_type
    """
    # Only want to watch directories that actually exist...
    paths = [path for path in paths if os.path.exists(path) and os.path.isdir(path)]

    if len(paths) == 0:
        raise ValueError, "Must supply at least one directory path"
    if action is None:
        raise ValueError, "Must supply event handler callback"

    # Start listening...
    monitor = dirwatch.create_dirwatcher()
    monitor.events.subscribe(EV_FILE_CREATED, action)
    try:
        for path in paths:
            print "Adding watch:", path
            monitor.add_watch(path)
        
        monitor.start_monitor()  # non-blocking
        try:
            print "Listening (press Ctrl-C to quit)..."
            try:
                # dumb busy wait. 
                # The monitor itself will happily sit in the
                # background now, and we could be doing other
                # things here...
                while True:
                    time.sleep(0.5)
            except KeyboardInterrupt:
                print "quitting..."
        finally:
            monitor.stop_monitor()
    finally:
        monitor.events.unsubscribe(EV_FILE_CREATED, action)

    print "done."

if __name__ == "__main__":
    args = parse_cmdline()

    # If the user doesn't specify, we'll default to the current dir.
    if len(args.directories) == 0:
        args.directories.append(os.getcwd())

    monitor_directories(args.directories, action=on_new_file)
