import Queue
from eventmanager import EventManager
import eventtype
import os.path
import threading
import logging
import time

LOG_NAME = "dirwatch"
LOG_STANDARD_FMT = "%(asctime)s p=%(process)d t=%(thread)d [%(levelname)s] %(message)s\n"
_LOG_HANDLER_ADDED = False

class DirWatcher(object):
    """ Abstract base class.
    Provides core DirWatcher event handling, startup and shutdown interfaces.
    Actual work, like monitoring for file events, should be done by derivative classes.

    It is expected that the derivative classes will pass file events it sees to 
    the monitor_events queue. Each of these events is expected to appear in the form of
    a tuple-like object with at least 3 fields:

    (timestamp, path, event_id)

    Where event_id is one of the supported 'EV_*' event types in the eventtype module.

    These events will be dispatched from the queue in order of arrival to any interested
    listeners, via the EventManager.

    Interested listeners subscribe by registering a callable, and the event_id(s) they want,
    with the EventManager, self.events.
    """
    def __init__(self):
        self.events = EventManager([event_name for event_name in dir(eventtype) 
                                               if event_name.startswith("EV_")])
        self._notify_lock = threading.RLock()
        self.paths = []
        self.monitor_events = Queue.Queue()
        self._exit_notification = object()
        self.monitoring = False      
        self.init_logging(verbosity=logging.WARNING) 
        self._worker_thread = None
        self.snapshot = {}
        
    def init_logging(self, verbosity=logging.DEBUG):
        self.log = logging.getLogger(LOG_NAME)
        self.log.setLevel(verbosity)
        
        global _LOG_HANDLER_ADDED
        if not _LOG_HANDLER_ADDED:
            # log format
            fmt = logging.Formatter(LOG_STANDARD_FMT)
            
            # console output
            console = logging.StreamHandler()
            #console.setLevel(verbosity)
            console.setFormatter(fmt)
            
            self.log.addHandler(console)
            _LOG_HANDLER_ADDED = True
        
    def _process_incoming_queue(self):
        """ Core event dispatcher.

        Receives filesystem events from the monitor(s), and
        broadcasts them to interested listeners via the EventManager.
        """
        while self.monitoring:
            event = self.monitor_events.get()
            
            if event is self._exit_notification:
                self.log.debug("_process_incoming_queue: EXIT_NOTIFICATION received")
                return
            
            # Otherwise, expecting events...
            try:
                timestamp, fullpath, event_id = event
                with self._notify_lock:
                    self.events.notify(event_id, path=fullpath, timestamp=timestamp)
            except Exception, details:
                self.log.error("_process_incoming_queue: Exception: %s", details)   
            
            self.monitor_events.task_done()                             
        
    def add_watch(self, path):
        """ Adds the given path to the watch list. 
        
        NOTE: If monitoring is already active, this change won't take effect
          until it is restarted.
        """
        if (path not in self.paths) and os.path.exists(path) and os.path.isdir(path):
            self.paths.append(path)
            
            self.log.debug("add_watch: getting initial snapshot of %s", path)
            self.snapshot[path] = self.get_dir_snapshot(path)            
            self.log.debug("add_watch: done.")
            
            with self._notify_lock:
                self.events.notify(eventtype.EV_WATCH_ADDED, path=path)
            return True
        return False
            
    def start_monitor(self):
        """  Begin watching the directories, given in self.paths, for events.
        
        Subsequent calls, if the montior is already active, will be ignored, and 
        return False.
        
        Returns:
            True, if the monitor starts successfully.

        NOTE: This base-class routine simply sets some state, and broadcasts
            a eventtype.EV_MONITOR_STARTED message. The descendant classes
            are expected to then start whatever monitoring mechanism they have
            provided.
        """
        if not self.monitoring:
            assert self._worker_thread is None
            
            self.monitoring = True
            self._worker_thread = threading.Thread(target=self._process_incoming_queue)
            self._worker_thread.start()
            
            with self._notify_lock:
                self.events.notify(eventtype.EV_MONITOR_STARTED)
    
    def stop_monitor(self):
        """ Stop the event monitor.

        NOTE: This base-class routine simply resets the state (set by start_monitor),
            and broadcasts a eventtype.EV_MONITOR_STOPPED message.
            Descendant is expected to cleanly shutdown its monitoring mechanism.
        """
        if self.monitoring:
            self.monitor_events.put(self._exit_notification)
            self.monitoring = False
            self._worker_thread.join()   
            with self._notify_lock:
                self.events.notify(eventtype.EV_MONITOR_STOPPED) 
                
    def get_dir_snapshot(self, directory, include_subdirs=False):
        """ Builds a set of filepaths in the given directory.
        
        Intension is to be able to find before/after differences, when
        there are too many events to process for our monitor, and we need
        to refresh our view to catch up.
        """
        def _iter_paths():
            for path, subdirs, filenames in os.walk(directory):
                if (not include_subdirs) and (path != directory):
                    break
                for filename in filenames:
                    yield os.path.join(path, filename)
        return set(_iter_paths())
    
    def generate_dir_snapshot_diff_events(self, before, after):
        """ Given two dir snapshot sets, send the add/delete events for the file lists. """
        additions = after - before
        for filepath in additions:
            with self._notify_lock:
                self.events.notify(eventtype.EV_FILE_CREATED, path=filepath, timestamp=time.time())
        additions = None 
        
        removals = before - after
        for filepath in removals:
            with self._notify_lock:
                self.events.notify(eventtype.EV_FILE_DELETED, path=filepath, timestamp=time.time())
        removals = None
        
