import time

class Event(object):
  def __init__(self):
    self.timestamp = time.time()
    self.event_type = None


class SubscriberList(object):
  def __init__(self, event_type):
    self._subscribers = []
    self.event_type = event_type
    
  def _build_event(self, attributes=None):
    """ Event object factory.

    If an attributes dictionary is given, the keys & values
    will be added to the object as attributes.
    """
    event = Event()
    event.event_type = self.event_type

    if attributes is not None:
      for key, value in attributes.iteritems():
        setattr(event, key, value)

    return event
    
  def subscribe(self, subscriber):
    """ Adds the subscriber to the notification list.
    
    Expects that subscriber is callable, and accepts at 
    least one argument (the event instance).
    """
    assert callable(subscriber), "subscriber should be callable"

    if subscriber not in self._subscribers:
      self._subscribers.append(subscriber)

  def unsubscribe(self, subscriber):
    self._subscribers.remove(subscriber)

  def notify(self, **kwargs):
    event = self._build_event(kwargs)
    for subscriber in self._subscribers:
      subscriber(event)
      
  def __iter__(self):
    return iter(self._subscribers)
    
  def __len__(self):
    return len(self._subscribers)


class EventManager(object):

  def __init__(self, event_types):

    self._listeners = {}
    for event_type in event_types:
      self._listeners[event_type] = SubscriberList(event_type)

  def __getitem__(self, event_type):
    return self._listeners[event_type]

  def __getattr__(self, event_type):
    return self[event_type]

  def get_event_types(self):
    return self._listeners.keys()

  def subscribe(self, event_type, subscriber):
    if not (subscriber in self[event_type]):
      self[event_type].subscribe(subscriber)

  def unsubscribe(self, event_type, subscriber):
    self[event_type].unsubscribe(subscriber)

  def notify(self, event_type, **kwargs):
    """ Notifies all listeners subscribed to the given event type.

    If any keyword arguments are supplied, they will be
    added to the event object as attributes.

    The event object will have at least one attribute named
    'event_type' -- which will be the event type being
    notified.
    """
    self[event_type].notify(**kwargs)
