from base import DirWatcher, LOG_STANDARD_FMT

import sys
import os
import threading
import logging
import logging.handlers
import time
import eventtype

if os.name == "nt":
    try:
        import pywintypes
        import win32api
        import win32file
        import win32event
        import win32con
        from win32con import FILE_SHARE_READ, FILE_SHARE_WRITE, FILE_SHARE_DELETE, \
                             OPEN_EXISTING, FILE_FLAG_BACKUP_SEMANTICS, FILE_FLAG_OVERLAPPED, \
                             FILE_NOTIFY_CHANGE_FILE_NAME, FILE_NOTIFY_CHANGE_LAST_WRITE
        from ntsecuritycon import FILE_LIST_DIRECTORY 
        from winerror import ERROR_SUCCESS, ERROR_ABANDONED_WAIT_0                                  
    except ImportError:
        print >> sys.stderr, "pywin32 libraries need to be installed."
        print >> sys.stderr, "See: https://sourceforge.net/projects/pywin32/"
        raise
    
def get_cpu_count():
    sysinfo = win32api.GetSystemInfo()
    dwNumberOfProcessors = 5 # 5th member of the SYSTEM_INFO struct.
    return sysinfo[dwNumberOfProcessors]

class DirWatcherNT(DirWatcher):
    """ Windows implementation of DirWatcher.

    Relies on ReadDirectoryChangesW(). Each watch request is assigned
    to a IO completion port (IOCP). 

    A thread pool waits on the IOCP, and when changes occur, the IOCP signals.
    One of the worker threads will parse the relevant result buffer, enqueuing the 
    events to the DirWatcher.monitor_events queue.

    This class attempts recovery in the event of ReadDirectoryChangesW() buffer 
    overflows.
    """
    
    READ_BUFFER_SIZE = 8192 
    EXIT_NOTIFICATION = -1
    
    DIRINFO_HANDLE_IDX = 0
    DIRINFO_PATH_IDX = 1
    DIRINFO_BUF_IDX = 2
    DIRINFO_OV_IDX = 3
    
    def __init__(self):
        super(DirWatcherNT, self).__init__()
        self.started = False
        self.dir_handles = {}
        self._listener_threads = []
        self._state_lock = threading.RLock()
        self._recover_lock = threading.RLock()
        
        self._iocp = None
        
                
    def start_monitor(self):
        """ Begin watching the directories, given in self.paths, for events.
        
        Subsequent calls, if the montior is already active, will be ignored, and 
        return False.
        
        Returns:
            True, if the monitor starts successfully.
        """
        super(DirWatcherNT, self).start_monitor()
        
        if len(self.paths) == 0:
            return False
        if self.started:
            return False
                
        # Start shooting off the watches...
        for path in self.paths:
            hdir = self._open_dir(path)
            self._assoc_with_iocp(hdir, path)
            self._send_dir_watch_request(hdir)
            
            
    def stop_monitor(self):
        """ Stops event monitoring.
        
        Waits for worker threads to shutdown, and flushes event queue.
        """
        # Shutdown the thread pool.
        self.log.debug("stop_monitor: Shutting down listeners")
        
        for n in range(len(self._listener_threads)):
            self._send_stop_watching_request()
        for thread in self._listener_threads:
            thread.join()
            
        # flush any stragglers... 
        # We don't need to process -- we're quitting.
        # Just don't want resource leaks at this point.
        if self._iocp:
            err = ERROR_SUCCESS
            while err == ERROR_SUCCESS: 
                err, numOfBytes, key, overlapped = win32file.GetQueuedCompletionStatus(self._iocp, 0)
        
        # Close our handles...
        for dirinfo in self.dir_handles.itervalues():
            try:
                dirinfo[self.DIRINFO_HANDLE_IDX].close()
            except:
                pass # already closed, so ignore.
        self.dir_handles = {}
        
        super(DirWatcherNT, self).stop_monitor()
        
            
    def _open_dir(self, path):
        """ Opens a handle to the given directory.
        
        Opens with shared read/write/delete, and access flags allowing use by 
        ReadDirectoryChangesW().
        """
        if os.path.exists(path) and os.path.isdir(path):
            file_handle = win32file.CreateFile(path,      # pointer to the file name
                                               FILE_LIST_DIRECTORY,    # access (read/write) mode
                                               FILE_SHARE_READ         # share mode
                                               | FILE_SHARE_WRITE
                                               | FILE_SHARE_DELETE,
                                               None,  # security descriptor
                                               OPEN_EXISTING,         # how to create
                                               FILE_FLAG_BACKUP_SEMANTICS # file attributes
                                               | FILE_FLAG_OVERLAPPED,
                                               None)
            if file_handle != win32file.INVALID_HANDLE_VALUE:
                return file_handle
        return None    
    
    
    def _recover_file_notify_error(self, path):
        """ Rebuilds internal view of watched directories, in the event of ReadDirectoryChangesW() failure.
        
        This normally happens when too many events have occurred before ReadDirectoryChangesW 
        manages to notify its listener.  ReadDirectoryChangesW ends up causing a buffer overflow, and
        we may miss many events as these, and the events that didn't fit into the result buffer, are lost.
        
        Short of reading the NTFS change journal, or building our own file system filter driver,
        our main option of recovery is to take another snapshot of the directory contents as they are now.
        
        Our base class' :generate_dir_snapshot_diff_events: will take care of sending "fake"
        events to simulate the creation of new files, that we hadn't seen previously, and the 
        deletion of files that we see are now missing.
        """
        self.log.debug("_recover_file_notify_error: Trying to recover from FILE_NOTIFY_INFORMATION error with new snapshot")
        with self._recover_lock:
            self.log.debug("_recover_file_notify_error: Got recover lock")
            new_snapshot = self.get_dir_snapshot(path)
            old_snapshot = self.snapshot.get(path, set([]))
            self.generate_dir_snapshot_diff_events(old_snapshot, new_snapshot)
            self.snapshot[path] = new_snapshot
            self.log.debug("_recover_file_notify_error: recovered.")
    
    
    def _listener_thread(self):
        """ Worker thread main routine. 
        
        Waits on IOCP for incoming events. When the event(s) arrive, these
        are parsed, and dirwatch event notifications are sent to subscribed 
        client listeners. 
        """
        while self._iocp:
            err, num_of_bytes, key, overlapped = win32file.GetQueuedCompletionStatus(self._iocp, -1)
            if err != ERROR_SUCCESS:
                self.log.debug("_listener_thread: GetQueuedCompletionStatus failed: %d, %s" % (err, win32api.FormatMessage(err)))
                if err == ERROR_ABANDONED_WAIT_0:
                    self.log.debug("_listener_thread: GetQueuedCompletionStatus: wait abandoned")
                    return
                
            if key == self.EXIT_NOTIFICATION:
                self.log.debug("_listener_thread: EXIT_NOTIFICATION received")
                return 
            
            with self._state_lock: 
                dirinfo = self.dir_handles.get(key, None)
            if not dirinfo:
                continue # back to sleep...
            try:
                path = dirinfo[self.DIRINFO_PATH_IDX]
                self.log.debug("_listener_thread: event received for '%s'", path)
                
                if num_of_bytes:
                    if num_of_bytes > self.READ_BUFFER_SIZE:
                        self.log.debug("_listener_thread: Received %d bytes - falling back to snapshots", num_of_bytes)
                        self._recover_file_notify_error(path)
                        continue
                    
                    try:
                        info = win32file.FILE_NOTIFY_INFORMATION(dirinfo[self.DIRINFO_BUF_IDX], num_of_bytes)
                        self.log.debug("_listener_thread: notify info: %s" % str(info))
                    except RuntimeError, err:
                        self.log.error("_listener_thread: failed to parse incoming FILE_NOTIFY_INFORMATION: %s", err)
                        self._recover_file_notify_error(path)              
                        info = None
                        
                    if not info:
                        continue
                    
                    # Need to translate the windows-specific action codes to dirwatch events.               
                    event_id = None 
                    for action, filename in info:
                        fullpath = os.path.join(path, filename)
                        
                        # FIXME: file rename/move ops aren't too easy with the given OS events.
                        #   We *could* check hashes on similarly timed delete-create event pairs, 
                        #   and see that the new file is the same as the old (hence, a move), but
                        #   I think speed will suffer greatly, for not much benefit.
                        
                        if action == FILE_NOTIFY_CHANGE_FILE_NAME:
                            if os.path.exists(fullpath):
                                event_id = eventtype.EV_FILE_CREATED
                            else:
                                event_id = eventtype.EV_FILE_DELETED
                                
                        elif action == FILE_NOTIFY_CHANGE_LAST_WRITE:
                            event_id = eventtype.EV_FILE_DATA_CHANGED
                        
                        if event_id:              
                            # send to our central queue....
                            # Sending the time, the fully qualified path, and the event type.
                            event = (time.time(), fullpath, event_id)  
                            self.log.debug("_listener_thread: enqueuing(%s)", str(event))    
                            self.monitor_events.put(event)
            finally:                    
                self._send_dir_watch_request(dirinfo[self.DIRINFO_HANDLE_IDX])         
        
        
    def _start_event_processor(self):  
        """ Spawns worker threads to listen on our IOCP for incoming events.
        
        Requires: Valid IOCP already exists.  Otherwise, a RuntimeError will be raised.
        """
        if self._iocp:
            listener_count = get_cpu_count() + 1 # Keep one ready when a busy thread sleeps...
            for lidx in range(listener_count):
                t = threading.Thread(target=self._listener_thread)
                self._listener_threads.append(t)
                t.start()
        else:
            raise RuntimeError, "IOCP not ready"

            
    def _assoc_with_iocp(self, handle, path):
        """ Start tracking this (handle,path) pair against incoming events from our IOCP.
        
        Associates the handle with the IO Completion Port, so future events against this handle
        will also be sent to our IOCP.
        
        Also adds an entry to our handle dictionary, so when an event arrives from the IOCP, 
        we can find the associated state (handle, path, result buffer...).
        """
        if not self._iocp:
            existing = None
        else:
            existing = self._iocp
            
        if handle:
            self.log.debug("_send_dir_watch_request: adding '%s'", path)
            
            # For each handle we track, we'll need a memory buffer to pass to ReadDirChanges, and an OVERLAPPED
            # structure. We'll keep these to the handle, so we don't lose their ref counts.
            buffer = win32file.AllocateReadBuffer(self.READ_BUFFER_SIZE)        
            overlapped = pywintypes.OVERLAPPED() 
                    
            with self._state_lock:
                # Need to ensure only our thread is working on the dir_handle dict here, as 
                # we're altering it.
                self.dir_handles[int(handle)] = [handle, path, buffer, overlapped]
            
            self._iocp = win32file.CreateIoCompletionPort(handle, existing, int(handle), 0)
            if not existing:
                # Ok - we have an IOCP to wait on now, start the thread pool.
                self._start_event_processor()
            else:
                assert existing is self._iocp, "Associating a new handle with existing IOCP should not change the IOCP."       

                
    def _send_dir_watch_request(self, hdir):  
        """ Submits the given directory handle to the OS event notification system.
        
        Non-blocking. Sends the given handle to ``ReadDirectoryChangesW()``.
        Events will be broadcast to us via our IOCP. 
        
        Note that this is a 'one-shot' notification. To continue monitoring this 
        directory, the handle will need to be resubmitted on receipt of the events.
        """
        if not self._iocp:
            return
        
        dirinfo = self.dir_handles[int(hdir)] # should be there by now...
        
        buffer = dirinfo[self.DIRINFO_BUF_IDX]
        overlapped = dirinfo[self.DIRINFO_OV_IDX]
        
        assert hdir is dirinfo[self.DIRINFO_HANDLE_IDX]        
                
        flags = FILE_NOTIFY_CHANGE_FILE_NAME | FILE_NOTIFY_CHANGE_LAST_WRITE
                
        # Since we're suppling an OVERLAPPED struct, this call shouldn't block.
        # Any notification events will get picked up by our threads listening on our
        # IO Completion Port.
        win32file.ReadDirectoryChangesW(hdir, buffer, False, flags, overlapped)                
    
    
    def _send_stop_watching_request(self):
        """ Sends exit notifications to the IOCP.
        
        One for each listening worker thread. Once the workers see these, they should
        exit on their own.
        """
        if self._iocp and len(self._listener_threads):
            win32file.PostQueuedCompletionStatus(self._iocp, 0, self.EXIT_NOTIFICATION, None)
    
    
    
def create_dirwatcher():
    """ Public factory for DirWatcherNT instances.
    
    Imported in the dirwatch module."""
    return DirWatcherNT()
